from typing import List
from ..interface import DynamicLightConfig
from ..opcutil import ColorData, get_color, spread, even_spread, rotate_right


class Scroll(DynamicLightConfig):
    """
    Scroll through a multi-colored line.
    """
    speed: int = 8
    color_list: List[ColorData]  # colors to scroll
    width: int  # (optional) number of pixels per color
    pixels: List[ColorData]  # current list of pixels

    def __init__(self, color_list: List[str], width: int = None, **kwargs):
        """
        Initialize a new Scroll configuration.

        :param color_list: the colors to use ("#RRGGBB" format)
        :param width:
        """
        super().__init__(**kwargs)
        self.validate_color_list(color_list)

        self.color_list = [get_color(c) for c in color_list]
        self.width = width

        if width:
            self.pixels = spread(self.color_list, width, self.num_leds)
        else:
            self.pixels = even_spread(self.color_list, self.num_leds)

    def __next__(self) -> List[ColorData]:
        self.pixels = rotate_right(self.pixels, 1)
        return self.pixels



from ...interface import LightConfig, DynamicLightConfig
from ..solid_color import SolidColor


class Strobe(DynamicLightConfig):
    """
    Add a strobe effect to a given ``LightConfig``.
    """
    _strobe_count = 0

    def __init__(self, config: LightConfig = SolidColor('#FFFFFF'),
                 strobe_speed: int = 2, **kwargs):
        """
        Initialize a new Stobe configuration.

        Please note that ``strobe_speed``
        uses different units than the ``speed`` parameter of
        ``DynamicLightConfig``. ``strobe_speed`` is defined as the number of
        "on" frames to push before turning the lights off for a frame. For
        example, say ``config`` is a light config that switches between red
        and green every frame and ``strobe_speed`` is set to 2. The animation
        would go like so:

        Frame 1: red
        Frame 2: green
        Frame 3: off (would be red)
        Frame 4: green
        Frame 5: red
        Frame 6: off (would be green)

        At a high level, the important thing to not is that a HIGHER
        ``strobe_speed`` value means a SLOWER strobe. A value of 1 means that
        every other frame is turned off, which is the maximum strobe speed.

        :param config: the light config to add a strobe effect to
        :param strobe_speed: the speed to strobe at (1 is fastest)
        """
        if isinstance(config, DynamicLightConfig):
            super().__init__(config.speed, config.num_leds, **kwargs)
        else:
            # if there is no speed to reference, use default value
            super().__init__(10, config.num_leds, **kwargs)

        self._config = config
        self.strobe_speed = strobe_speed

    def __next__(self):
        self._strobe_count = self._strobe_count + 1 % self.strobe_speed
        pixels = next(self._config)

        if self._strobe_count == 0:
            return [(0, 0, 0)] * self.num_leds
        else:
            return pixels



import os
import sys
import subprocess
import socket
import asyncio
import atexit
import logging

__all__ = ['FadecandyServer']

logger = logging.getLogger('fcserver')
logger.setLevel(logging.INFO)


class FadecandyServer:
    """
    Controller for Fadecandy server.
    """
    # TODO: Allow user to indicate where to redirect console output

    _server_running: bool = False
    _fcserver_proc: subprocess.Popen = None

    def start(self) -> None:
        """
        Run the Fadecandy server. Terminates on program exit.
        """

        async def _go():
            args = []
            if sys.platform == 'win32':
                server = 'fcserver.exe'
            elif sys.platform == 'darwin':
                server = 'fcserver-osx'
            else:
                server = 'fcserver-rpi'
                args.append('sudo')

            here = os.path.dirname(os.path.abspath(__file__))
            args.append(here + '/bin/' + server)

            _fcserver_proc = subprocess.Popen(args)
            logger.info(f'Started {server}')
            return _fcserver_proc

        if not self._server_running:
            # check if other instance of fcserver is running on port
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                result = s.connect_ex(('127.0.0.1', 7890))

            if result in {10061, 111}:  # nothing running
                self._fcserver_proc = asyncio.run(_go())
                self._server_running = True
                atexit.register(self.stop)  # stop fcserver on exit
            else:
                logger.info('Another instance of fcserver is already running')

    def stop(self) -> None:
        """
        Stop the Fadecandy server.
        """
        if self._server_running:
            self._fcserver_proc.terminate()
            self._server_running = False
            logger.info('Stopped fcserver')

    def __repr__(self):
        if self._server_running:
            end = ' RUNNING'
        else:
            end = ' STOPPED'
        return super().__repr__() + end



from typing import List

from ..interface import StaticLightConfig
from ..opcutil import ColorHex, ColorData, get_color


class SolidColor(StaticLightConfig):
    """
    Display a solid color.
    """
    color: ColorData

    def __init__(self, color: ColorHex, **kwargs):
        """
        Initialize a new SolidColor configuration.
        :param color: the color to dislpay ("#RRGGBB" format)
        """
        super().__init__(**kwargs)
        self.validate_color(color)

        self.color: ColorData = get_color(color)

    def pattern(self) -> List[ColorData]:
        return [self.color] * self.num_leds



#!/usr/bin/env python

"""Python Client library for Open Pixel Control
http://github.com/zestyping/openpixelcontrol

Sends pixel values to an Open Pixel Control server to be displayed.
http://openpixelcontrol.org/

Recommended use:

    import opc

    # Create a client object
    client = opc.Client('localhost:7890')

    # Test if it can connect (optional)
    if client.can_connect():
        print('connected to %s' % ADDRESS)
    else:
        # We could exit here, but instead let's just print a warning
        # and then keep trying to send pixels in case the server
        # appears later
        print('WARNING: could not connect to %s' % ADDRESS)

    # Send pixels forever at 30 frames per second
    while True:
        my_pixels = [(255, 0, 0), (0, 255, 0), (0, 0, 255)]
        if client.put_pixels(my_pixels, channel=0):
            print('...')
        else:
            print('not connected')
        time.sleep(1/30.0)

"""

import socket
import struct
import sys


class Client(object):

    def __init__(self, server_ip_port, long_connection=True, verbose=False):
        """Create an OPC client object which sends pixels to an OPC server.

        server_ip_port should be an ip:port or hostname:port as a single string.
        For example: '127.0.0.1:7890' or 'localhost:7890'

        There are two connection modes:
        * In long connection mode, we try to maintain a single long-lived
          connection to the server.  If that connection is lost we will try to
          create a new one whenever put_pixels is called.  This mode is best
          when there's high latency or very high framerates.
        * In short connection mode, we open a connection when it's needed and
          close it immediately after.  This means creating a connection for each
          call to put_pixels. Keeping the connection usually closed makes it
          possible for others to also connect to the server.

        A connection is not established during __init__.  To check if a
        connection will succeed, use can_connect().

        If verbose is True, the client will print debugging info to the console.

        """
        self.verbose = verbose

        self._long_connection = long_connection

        self._ip, self._port = server_ip_port.split(':')
        self._port = int(self._port)

        self._socket = None  # will be None when we're not connected

    def _debug(self, m):
        if self.verbose:
            print('    %s' % str(m))

    def _ensure_connected(self):
        """Set up a connection if one doesn't already exist.

        Return True on success or False on failure.

        """
        if self._socket:
            self._debug('_ensure_connected: already connected, doing nothing')
            return True

        try:
            self._debug('_ensure_connected: trying to connect...')
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect((self._ip, self._port))
            self._debug('_ensure_connected:    ...success')
            return True
        except socket.error:
            self._debug('_ensure_connected:    ...failure')
            self._socket = None
            return False

    def disconnect(self):
        """Drop the connection to the server, if there is one."""
        self._debug('disconnecting')
        if self._socket:
            self._socket.close()
        self._socket = None

    def can_connect(self):
        """Try to connect to the server.

        Return True on success or False on failure.

        If in long connection mode, this connection will be kept and re-used for
        subsequent put_pixels calls.

        """
        success = self._ensure_connected()
        if not self._long_connection:
            self.disconnect()
        return success

    def put_pixels(self, pixels, channel=0):
        """Send the list of pixel colors to the OPC server on the given channel.

        channel: Which strand of lights to send the pixel colors to.
            Must be an int in the range 0-255 inclusive.
            0 is a special value which means "all channels".

        pixels: A list of 3-tuples representing rgb colors.
            Each value in the tuple should be in the range 0-255 inclusive. 
            For example: [(255, 255, 255), (0, 0, 0), (127, 0, 0)]
            Floats will be rounded down to integers.
            Values outside the legal range will be clamped.

        Will establish a connection to the server as needed.

        On successful transmission of pixels, return True.
        On failure (bad connection), return False.

        The list of pixel colors will be applied to the LED string starting
        with the first LED.  It's not possible to send a color just to one
        LED at a time (unless it's the first one).

        """
        self._debug('put_pixels: connecting')
        is_connected = self._ensure_connected()
        if not is_connected:
            self._debug('put_pixels: not connected.  ignoring these pixels.')
            return False

        # build OPC message
        len_hi_byte = int(len(pixels)*3 / 256)
        len_lo_byte = (len(pixels)*3) % 256
        command = 0  # set pixel colors from openpixelcontrol.org

        header = struct.pack("BBBB", channel, command, len_hi_byte, len_lo_byte)

        pieces = [ struct.pack( "BBB",
                     min(255, max(0, int(r))),
                     min(255, max(0, int(g))),
                     min(255, max(0, int(b)))) for r, g, b in pixels ]

        if sys.version_info[0] == 3:
            # bytes!
            message = header + b''.join(pieces)
        else:
            # strings!
            message = header + ''.join(pieces)

        self._debug('put_pixels: sending pixels to server')
        try:
            self._socket.send(message)
        except socket.error:
            self._debug('put_pixels: connection lost.  could not send pixels.')
            self._socket = None
            return False

        if not self._long_connection:
            self._debug('put_pixels: disconnecting')
            self.disconnect()

        return True

    def set_interpolation(self, enabled = True):
        """
        Enables or disables frame interpolation on runtime.

        Return True on success.
        """
        self._debug('set_interpolation: connecting')
        is_connected = self._ensure_connected()
        if not is_connected:
            self._debug('set_interpolation: not connected.  ignoring reconfiguration.')
            return False
    
        #build firmaware configuration message as documented on
        #https://github.com/scanlime/fadecandy/blob/master/doc/fc_protocol_opc.md#set-firmware-configuration
        if enabled:
            config_bit = 0
        else:
            config_bit = 2
        message = struct.pack('BBBBBBBBB', 0, 255, 0, 5, 0, 1, 0, 2, config_bit)
    
        self._debug('set_interpolation: sending firmware configuration')
        try:
            self._socket.send(message)
        except socket.error:
            self._debug('set_interpolation: connection lost.  could not send firmware configuration.')
            self._socket = None
            return False
        if not self._long_connection:
            self._debug('set_interpolation: disconnecting')
            self.disconnect()
        return True





"""
Module for Open Pixel Control utilities. This module provides color definitions
and list manipulation functions. The list manipulation functions are designed
to be used with lists of ColorData (for pushing colors to LEDs), but they can
be used with any kinds of lists.
"""

import re
import math

from typing import Any, Tuple, List

__all__ = [
    'ColorData',
    'ColorHex',
    'ColorList',
    'is_color',
    'is_color_list',
    'get_color',
    'shift',
    'even_spread',
    'spread',
    'rotate_left',
    'rotate_right'
]


ColorData = Tuple[float, float, float]  # 3-tuple representing actual RGB value
ColorHex = str  # ``color`` must be "#RRGGBB" format
ColorList = List[ColorHex]  # ``color_list`` must be non-empty list of ColorHex


def is_color(v: Any) -> bool:
    """
    Determine whether ``v`` is a valid ``ColorHex``.
    """
    return isinstance(v, str) and bool(re.match(r'^#[A-Fa-f0-9]{6}$', v))


def is_color_list(v: Any) -> bool:
    """
    Determine whether ``v`` is a valid ``ColorList``.
    """
    return isinstance(v, list) and len(v) > 0 and all([is_color(c) for c in v])


def get_color(hex_str: ColorHex) -> ColorData:
    """
    Calculate a 3-tuple representing the color in the given hex.

    :param hex_str: the desired color in the format #RRGGBB.
    :return: a 3-tuple of RGB values ``hex_str`` represents
    :raises ValueError: if ``hex_str`` is improperly formatted
    """
    if not is_color(hex_str):
        hex_repr = f"'{hex_str}'" if type(hex_str) is str else str(hex_str)
        raise ValueError("Please provide a color in the format '#RRGGBB'. "
                         f"Received {hex_repr}.")

    def hexfloat(h: str) -> float:
        return float(int(h, 16))

    red = hexfloat(hex_str[1:3])
    blue = hexfloat(hex_str[3:5])
    green = hexfloat(hex_str[5:7])

    return red, blue, green


def shift(current: ColorData, goal: ColorData, p: float) -> ColorData:
    """
    Shift a color towards another.

    :param current: a 3-tuple representing the starting color
    :param goal: a 3-tuple representing the color to shift towards
    :param p: a value indicating how far to shift (0 => no shift, 1 => ``goal``)
    :return: the shifted color
    """
    new_vals = [c + ((g - c) * p) for c, g in zip(current, goal)]
    # explicitly give the first 3 values so it doesn't complain about being
    # Tuple[float, ...] rather than Tuple[float, float, float]
    return new_vals[0], new_vals[1], new_vals[2]


def even_spread(vals: List, n: int) -> List:
    """
    Spread out the values across a larger list in order as evenly as possible.

    Example:
    >>> even_spread([1, 2, 3], 8) == [1, 1, 1, 2, 2, 2, 3, 3]

    :param vals: a list of values to spread
    :param n: the length of the list to create from the given values
    :return: a list consisting of ``colors`` spread across ``num_leds`` indices
    :raises ValueError: if ``colors`` is empty or ``num_leds`` is negative
    """
    if len(vals) < 1:
        raise ValueError('no values provided')
    if n < 0:
        raise ValueError('list length cannot be negative')

    new_vals = []
    length_per_val = math.floor(n / len(vals))
    remainder = n % len(vals)
    extra_vals = vals[:remainder]

    for val in vals:
        new_vals += [val] * length_per_val
        if val in extra_vals:
            new_vals.append(val)

    return new_vals


def spread(vals: List, seq_length: int, list_length: int):
    """
    Spread out values across a certain number of indices using the specified
    number of times to repeat each value, repeating if the end of ``vals`` is
    reached.

    Examples:
    >>> spread([1, 2, 2, 3], 2, 7) == [1, 1, 2, 2, 2, 2, 3]

    >>> spread([4, 5], 3, 8) == [4, 4, 4, 5, 5, 5, 4, 4]

    :param vals: the values to spread
    :param seq_length: the length of each value sequence
    :param list_length: the length of the list to create from the given colors
    :return: a list consisting of sequences of values in ``vals`` of length
        ``r`` each
    :raises ValueError: if ``vals`` is empty or either ``n`` or
        ``r`` is negative
    """
    if len(vals) < 1:
        raise ValueError('no values provided')
    if list_length < 0:
        raise ValueError('list length cannot be negative')
    if seq_length < 0:
        raise ValueError('sequence length cannot be negative')

    new_vals = []
    idx = 0
    remaining = list_length
    while remaining > 0:
        new_vals += [vals[idx]] * min(seq_length, remaining)
        idx = (idx + 1) % len(vals)
        remaining -= seq_length

    return new_vals


def rotate_left(l: List, n: int):
    """
    Generate a version of the given list rotated left.

    :param l: the list to rotate
    :param n: how many spaces to rotate the list
    :return: the rotated list
    """
    n %= len(l)
    return l[n:] + l[:n]


def rotate_right(l: List, n: int):
    """
    Generate a version of the given list rotated right

    :param l: the list to rotate
    :param n: how many spaces to rotate the list
    :return: the rotated list
    """
    n %= len(l)
    return l[-n:] + l[:-n]



"""
This package contains patterns which are modifiers for existing patterns. If
a pattern takes another pattern in its constructor, it belongs in this package.
"""

from .strobe import Strobe

__all__ = ['Strobe']



from typing import List
from ..opcutil import ColorHex, ColorData, get_color, spread, even_spread
from ..interface import StaticLightConfig


class Stripes(StaticLightConfig):
    """
    Display multiple static colors.
    """
    color_list: List[ColorData]  # colors to display
    width: int  # (optional) number of pixels per color

    def __init__(self, color_list: List[ColorHex], width: int = None, **kwargs):
        """
        Initialize a new Stripes configuration.
        :param color_list: the colors to use ("#RRGGBB" format)
        :param width: the width of each color strip
        """
        super().__init__(**kwargs)
        self.validate_color_list(color_list)

        self.color_list = [get_color(c) for c in color_list]
        self.width = width

    def pattern(self) -> List[ColorData]:
        if self.width:
            return spread(self.color_list, self.width, self.num_leds)
        else:
            return even_spread(self.color_list, self.num_leds)



from . import modifiers

from .fade import Fade
from .scroll import Scroll
from .solid_color import SolidColor
from .stripes import Stripes
from .off import Off

__all__ = ['Fade', 'Scroll', 'SolidColor', 'Stripes', 'Off']



from .solid_color import SolidColor


class Off(SolidColor):
    """
    Turn the LEDs off.
    """

    def __init__(self, **kwargs):
        super().__init__('#000000', **kwargs)  # put black pixels



import abc
import time

from typing import List, Iterator
from . import opc
from .opcutil import ColorData, ColorHex, is_color, is_color_list

__all__ = ['LightConfig', 'DynamicLightConfig', 'StaticLightConfig']

# IDEA
# Lighting configurations are iterators that generate the next list of pixels
# to put onto an LED strip. The run method steps through the iterator and does
# the work of pushing the generated list to the Fadecandy client.


# TODO: Add ability to control brightness
class LightConfig(abc.ABC):
    """
    Abstract base class for an LED lighting configuration.
    """
    client: opc.Client

    def __init__(self, num_leds: int = 512, **kwargs):
        """
        Initialize a new LightConfig.

        :param num_leds: the number of LEDs
        """
        self.num_leds: int = num_leds

    def __iter__(self) -> Iterator:
        """
        Define any LightConfig to be iterable.
        """
        return self

    @abc.abstractmethod
    def __next__(self) -> List[ColorData]:
        """
        Get the next list of colors to push to the Fadecandy client.
        """

    @abc.abstractmethod
    def run(self, host: str = 'localhost', port: int = 7890) -> None:
        """
        Run this lighting configuration.

        :param host: hostname or IP address of Fadecandy server
        :param port: port that the Fadecandy server is running on
        """
        self.client = opc.Client(f'{host}:{port}')

    @staticmethod
    def factory(pattern: str, strobe: bool = False, color: str = None,
                color_list: List[str] = None,
                speed: int = None) -> 'LightConfig':
        """
        Generate a ``LightConfig`` based on keywaord arguments. Different
        patterns differ in required keyword arguments.

        :param pattern: the name of the desired lighting configuration
        :param strobe: whether to add a strobe effect
        :param color: the color to use (if applicable)
        :param color_list: the list of colors to use (if applicable)
        :param speed: the speed to run a dynamic configuration at
        :return: an instance of the class associated with ``pattern``
        :raises ValueError: if ``pattern`` is not associated with any patterns
            or the required arguments for the specified config are not provided
        """
        # importing patterns at the top of file causes circular import issues
        from . import patterns

        config = dict()
        if color:
            config['color'] = color
        if color_list:
            config['color_list'] = color_list
        if speed:
            config['speed'] = speed

        try:
            light = getattr(patterns, pattern)(**config)
        except AttributeError:
            raise ValueError(f'{pattern!r} is not associated with any lighting '
                             f'configurations')

        if strobe:
            light = patterns.modifiers.Strobe(light)

        return light

    @staticmethod
    def validate_color(color: ColorHex) -> None:
        """
        Ensure that a value passed as the ``color`` parameter is valid. This
        means it is a string of the format "#RRGGBB". If the given value
        is invalid, an exception is raised.

        This method should be called in the contructor of ``LightConfig``
        subclasses that take a ``color`` parameter.

        :param color: the ``color`` to validate
        :raises TypeError: if ``color`` is not valid
        """
        if not is_color(color):
            raise TypeError('color must be in format "#RRGGBB"')

    @staticmethod
    def validate_color_list(color_list: List[ColorHex]) -> None:
        """
        Ensure that a value passed as the ``color_list`` parameter is valid.
        This means it is a list of strings in the format "#RRGGBB" with a length
        of at least one. In other words, it is a non-empty ``list[ColorHex]``.
        If the given value is invalid, an exception is raised.

        This method should be called in the contructor of ``LightConfig``
        subclasses that take a ``color_list`` parameter.

        :param color_list: the ``color_list`` to validate
        :raises TypeError: if ``color_list`` is not valid
        """
        if not is_color_list(color_list):
            raise TypeError('color_list must be a non-empty list of strings '
                            'in format "#RRGGBB')


class StaticLightConfig(LightConfig, abc.ABC):
    """
    A lighting configuration that displays an unmoving pattern.
    """

    def __next__(self):
        return self.pattern()

    def run(self, host: str = 'localhost', port: int = 7890) -> None:
        super().run(host, port)  # initialize client
        black = [(0, 0, 0)] * self.num_leds
        pattern = self.pattern()

        # turn off LEDs
        self.client.put_pixels(black)
        self.client.put_pixels(black)

        # fade in to pattern
        time.sleep(0.5)
        self.client.put_pixels(pattern)

    @abc.abstractmethod
    def pattern(self) -> List[ColorData]:
        """
        Define the pattern this lighting configuration should display.
        :return: a list of RGB values to display
        """


class DynamicLightConfig(LightConfig, abc.ABC):
    """
    A lighting configuration that displays a moving pattern.
    """
    speed: int

    def __init__(self, speed: int = None, num_leds: int = 512, **kwargs):
        """
        Initialize a new DynamicLightConfig.

        :param speed: the speed at which the lights change (updates per second)
        :param num_leds: the number of LEDs
        """
        super().__init__(num_leds, **kwargs)
        if speed:
            self.speed = speed

    def run(self, host: str = 'localhost', port: int = 7890) -> None:
        super().run(host, port)  # initialize client
        while True:
            pixels = next(self)
            self.client.put_pixels(pixels)
            time.sleep(1 / self.speed)



from setuptools import setup, find_packages

with open('README.md') as readme:
    long_description = readme.read()

setup(
    name='opclib',
    version='0.0.6',
    author='Graham Preston',
    author_email='graham.preston@gmail.com',
    description='A wrapper library for Open Pixel Control.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/gcpreston/opclib',
    packages=find_packages(exclude=('opclib.tests',)),
    package_dir={'opclib': 'opclib'},
    package_data={'opclib': ['bin/fcserver*']},
    include_package_data=True,
    zip_safe=False,
    # python_requires='>=3.7',  # TODO: Test other Python versions
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    project_urls={
        'Source': 'https://github.com/gcpreston/opclib',
    }
)



from typing import List
from ..interface import DynamicLightConfig
from ..opcutil import ColorHex, ColorData, get_color, shift


class Fade(DynamicLightConfig):
    """
    Fade between specified colors.
    """
    speed: int = 5
    color_list: List[ColorData]  # colors to fade between
    pixels: List[ColorData]  # current list of pixels

    _current_color: ColorData
    _color_index = 0  # index of color being faded towards in self.colors
    _fade_index = 0  # how far we are between two colors [0-9]

    def __init__(self, color_list: List[ColorHex], **kwargs):
        """
        Initialize a new Fade configuration.
        :param color_list: the colors to use ("#RRGGBB" format)
        """
        super().__init__(**kwargs)
        self.validate_color_list(color_list)

        self.color_list = [get_color(c) for c in color_list]
        self._current_color = self.color_list[0]
        self.pixels = [self._current_color] * self.num_leds

    def __next__(self) -> List[ColorData]:
        if self._fade_index == 9:
            # go to next color
            self._color_index = (self._color_index + 1) % len(self.color_list)

        # increment _fade_index, or wrap back to 0
        self._fade_index = (self._fade_index + 1) % 10

        # shift pixels 10% towards the next color
        self._current_color = shift(self._current_color,
                                    self.color_list[self._color_index], 0.1)
        self.pixels = [self._current_color] * self.num_leds

        return self.pixels



from .interface import *
from .fcserver import *
from .opcutil import *
from .patterns import *

pattern_names = patterns.__all__
modifier_names = patterns.modifiers.__all__

__all__ = (
    interface.__all__
    + fcserver.__all__
    + opcutil.__all__
    + ['patterns', 'pattern_names', 'modifier_names']
)
